@@ -54,6 +54,8 @@
 #include <assert.h>
 #include <pwd.h>
 #include <grp.h>
+#include <limits.h>
+#include <signal.h>
 #ifdef HAVE_SETXATTR
 #include <sys/xattr.h>
 #endif
@@ -149,7 +151,6 @@ static int getattr_common(const char *path, struct stat *stbuf);
 /* Chowns a new file if necessary. */
 static void chown_new_file(const char *path, struct fuse_context *fc, int (*chown_func)(const char*, uid_t, gid_t));
 
-
 /* FUSE callbacks */
 static void *bindfs_init();
 static void bindfs_destroy(void *private_data);
@@ -188,12 +189,18 @@ static int bindfs_fsync(const char *path, int isdatasync,
 
 
 static void print_usage(const char *progname);
+
 static int process_option(void *data, const char *arg, int key,
                           struct fuse_args *outargs);
 static int parse_mirrored_users(char* mirror);
 static int parse_user_map(UserMap *map, UserMap *reverse_map, char *spec);
-static char* get_working_dir();
+static char *get_working_dir();
 static void maybe_stdout_stderr_to_file();
+
+/* Sets up handling of SIGUSR1. */
+static void setup_signal_handling();
+static void signal_handler(int sig);
+
 static void atexit_func();
 
 static int is_mirroring_enabled()
@@ -283,14 +290,26 @@ static int getattr_common(const char *procpath, struct stat *stbuf)
     return 0;
 }
 
+/* FIXME: another thread may race to see the old owner before the chown is done.
+          Is there a scenario where this compromises security? Or application correctness? */
 static void chown_new_file(const char *path, struct fuse_context *fc, int (*chown_func)(const char*, uid_t, gid_t))
 {
     uid_t file_owner;
     gid_t file_group;
-    
+
     if (settings.create_policy == CREATE_AS_USER) {
+        char *path_copy;
+        const char *dir_path;
+        struct stat stbuf;
+
         file_owner = fc->uid;
         file_group = fc->gid;
+
+        path_copy = strdup(path);
+        dir_path = my_dirname(path_copy);
+        if (lstat(dir_path, &stbuf) != -1 && stbuf.st_mode & S_ISGID)
+            file_group = -1;
+        free(path_copy);
     } else {
         file_owner = -1;
         file_group = -1;
@@ -311,6 +330,8 @@ static void chown_new_file(const char *path, struct fuse_context *fc, int (*chow
     }
 }
 
+
+
 static void *bindfs_init()
 {
     assert(settings.permchain != NULL);
@@ -399,10 +420,17 @@ static int bindfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
     struct dirent *de;
     struct stat st;
     int result = 0;
-    (void) path;
+    long pc_ret;
     
-    de_buf = malloc(offsetof(struct dirent, d_name) + pathconf(path, _PC_NAME_MAX) + 1);
+    path = process_path(path);
     
+    pc_ret = pathconf(path, _PC_NAME_MAX);
+    if (pc_ret < 0) {
+        DPRINTF("pathconf failed: %s (%d)", strerror(errno), errno);
+        pc_ret = NAME_MAX;
+    }
+    de_buf = malloc(offsetof(struct dirent, d_name) + pc_ret + 1);
+
     seekdir(dp, offset);
     while (1) {
         result = readdir_r(dp, de_buf, &de);
@@ -949,13 +977,14 @@ static void print_usage(const char *progname)
            "  --ctime-from-mtime        Read file properties' change time\n"
            "                            from file content modification time.\n"
            "  --hide-hard-links         Always report a hard link count of 1.\n"
+           "  --multithreaded           Enable multithreaded mode. See man page\n"
+           "                            for security issue with current implementation.\n"
            "\n"
            "FUSE options:\n"
            "  -o opt[,opt,...]          Mount options.\n"
            "  -r      -o ro             Mount strictly read-only.\n"
            "  -d      -o debug          Enable debug output (implies -f).\n"
            "  -f                        Foreground operation.\n"
-           "  -s                        Disable multithreaded operation.\n"
            "\n"
            "(*: root only)\n"
            "\n",
@@ -985,7 +1014,8 @@ enum OptionKey {
     OPTKEY_XATTR_READ_WRITE,
     OPTKEY_REALISTIC_PERMISSIONS,
     OPTKEY_CTIME_FROM_MTIME,
-    OPTKEY_HIDE_HARD_LINKS
+    OPTKEY_HIDE_HARD_LINKS,
+    OPTKEY_MULTITHREADED
 };
 
 static int process_option(void *data, const char *arg, int key,
@@ -1234,7 +1264,7 @@ static void maybe_stdout_stderr_to_file()
     strcat(path, "/");
     strcat(path, filename);
     
-    fd = open(path, O_CREAT | O_WRONLY);
+    fd = open(path, O_CREAT | O_WRONLY, 0666);
     free(path);
     
     fchmod(fd, 0777 & ~settings.original_umask);
@@ -1245,7 +1275,7 @@ static void maybe_stdout_stderr_to_file()
 #endif
 }
 
-static char* get_working_dir()
+static char *get_working_dir()
 {
     size_t buf_size = 4096;
     char* buf = malloc(buf_size);
@@ -1256,6 +1286,21 @@ static char* get_working_dir()
     return buf;
 }
 
+static void setup_signal_handling()
+{
+    struct sigaction sa;
+    sa.sa_handler = signal_handler;
+    sigemptyset(&sa.sa_mask);
+    sa.sa_flags = 0;
+    
+    sigaction(SIGUSR1, &sa, NULL);
+}
+
+static void signal_handler(int sig)
+{
+    invalidate_user_cache();
+}
+
 static void atexit_func()
 {
     free(settings.original_working_dir);
@@ -1290,6 +1335,7 @@ int main(int argc, char *argv[])
         char *create_for_group;
         char *create_with_perms;
         int no_allow_other;
+        int multithreaded;
     } od;
 
     #define OPT2(one, two, key) \
@@ -1334,6 +1380,7 @@ int main(int argc, char *argv[])
         OPT2("--realistic-permissions", "realistic-permissions", OPTKEY_REALISTIC_PERMISSIONS),
         OPT2("--ctime-from-mtime", "ctime-from-mtime", OPTKEY_CTIME_FROM_MTIME),
         OPT2("--hide-hard-links", "hide-hard-links", OPTKEY_HIDE_HARD_LINKS),
+        OPT_OFFSET2("--multithreaded", "multithreaded", multithreaded, -1),
         FUSE_OPT_END
     };
 
@@ -1458,6 +1505,11 @@ int main(int argc, char *argv[])
     }
 
 
+    /* Single-threaded mode by default */
+    if (!od.multithreaded) {
+        fuse_opt_add_arg(&args, "-s");
+    }
+
     /* Add default fuse options */
     if (!od.no_allow_other) {
         fuse_opt_add_arg(&args, "-oallow_other");
@@ -1499,6 +1551,9 @@ int main(int argc, char *argv[])
         bindfs_oper.removexattr = NULL;
     }
 
+    /* fuse_main will daemonize by fork()'ing. The signal handler will persist. */
+    setup_signal_handling();
+
     fuse_main_return = fuse_main(args.argc, args.argv, &bindfs_oper);
 
     fuse_opt_free_args(&args);
